#region License
/* Copyright (c) 2007 Dirk Schnorpfeil
 * Midi Multimedia Libraries Copyright (c) 2005 Leslie Sanford
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to 
 * deal in the Software without restriction, including without limitation the 
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
 * sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
 * THE SOFTWARE.
 */
#endregion

#region Contact
/*
 * Dirk Schnorpfeil
 * Email: D.Schnorpfeil@web.de
 */
#endregion

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Diagnostics;

namespace MC808Librarian
{
    class Patch
    {
        #region IComparer
        public class ByName : IComparer<Patch>
        {
            public int Compare(Patch lhs, Patch rhs)
            {
                return lhs.Name.CompareTo(rhs.Name);
            }
        }

        public class ByCatShort : IComparer<Patch>
        {
            public int Compare(Patch lhs, Patch rhs)
            {
                return lhs.CategoryShort.CompareTo(rhs.CategoryShort);
            }
        }

        public class ByCatLong : IComparer<Patch>
        {
            public int Compare(Patch lhs, Patch rhs)
            {
                return lhs.CategoryLong.CompareTo(rhs.CategoryLong);
            }
        }
        #endregion

        public Patch()
        {
        }

        public Patch(Patch rhs)
        {
            _common = (byte[])rhs._common.Clone();
            _tmt = (byte[])rhs._tmt.Clone();
            for (int tone = 0; tone < 4; tone++)
            {
                _tone[tone] = (byte[])rhs._tone[tone].Clone();
            }
        }

        #region Properties
        public string Name
        { 
            get 
            {
                StringBuilder s = new StringBuilder("            ");
                for (int i = 0; i < 12; i++)
                    s[i] = (char)_common[i];
                return s.ToString().TrimEnd(blanks);
            }
        }

        public int Category
        {
            get { return _common[0x0C]; }
        }

        public string CategoryShort
        {
            get
            {
                int cat = Category;
                if (cat >= 0 && cat < _categoryShort.GetLength(0))
                    return _categoryShort[cat];
                else
                    return "???";
            }
        }

        public string CategoryLong
        {
            get
            {
                int cat = Category;
                if (cat >= 0 && cat < _categoryLong.GetLength(0))
                    return _categoryLong[cat];
                else
                    return "???";
            }
        }

        public int WaveGroup(int tone)
        {
            int wave_group = (_tone[tone][0x28] << 12) | (_tone[tone][0x29] << 8) | (_tone[tone][0x2A] << 4) | (_tone[tone][0x2B]);
            return wave_group;
        }

        public int WaveLeft(int tone)
        {
            Debug.Assert(tone >= 0 && tone <= 3);
            int waveLeft = (_tone[tone][0x2C] << 12) | (_tone[tone][0x2D] << 8) | (_tone[tone][0x2E] << 4) | (_tone[tone][0x2F]);
            return waveLeft;
        }

        public int WaveRight(int tone)
        {
            Debug.Assert(tone >= 0 && tone <= 3);
            int wave_right = (_tone[tone][0x30] << 12) | (_tone[tone][0x31] << 8) | (_tone[tone][0x32] << 4) | (_tone[tone][0x33]);
            return wave_right;
        }
        public byte[] Common
        {
            get { return _common; }
        }

        public byte[] Tmt
        {
            get { return _tmt; }
        }

        public byte[] Tone(int tone)
        {
            return _tone[tone];
        }

        public int SysExCount
        {
            get { return _sysExCnt; }
        }
        #endregion

        #region File IO
        public void Read(BinaryReader reader)
        {
            _common = reader.ReadBytes(0x51);
            _tmt = reader.ReadBytes(0x29);
            for (int tone = 0; tone < 4; tone++)
            {
                _tone[tone] = reader.ReadBytes(0x8b);
            }
        }

        public void Write(BinaryWriter writer)
        {
            writer.Write(_common);
            writer.Write(_tmt);
            for (int tone = 0; tone < 4; tone++)
            {
                writer.Write(_tone[tone]);
            }
        }
        #endregion

        public Patch(byte[] common, byte[] tmt, byte[,] tones)
        {
            _common = (byte[])common.Clone();
            _tmt = (byte[])tmt.Clone();
            for (int tone = 0; tone < 4; tone++)
            {
                _tone[tone] = new byte[0x8B];
                for (int i = 0; i < 0x8B; i++)
                    _tone[tone][i] = tones[tone, i];
            }
        }

        public void AddSysex(byte[] msg)
        {
            // Todo: buffer links und rechts krzen und dann setzen
            _test[_sysExCnt] = msg;

            if (_sysExCnt == 0) // Common
            {
                _common = new byte[0x51];
                for (int n = 0; n < 0x51; n++)
                {
                    _common[n] = msg[n + 0x0b];
                }
            }
            else if (_sysExCnt == 1) // TMT
            {
                _tmt = new byte[0x29];
                for (int n = 0; n < 0x29; n++)
                {
                    _tmt[n] = msg[n + 0x0b];
                }
            }
            else if (_sysExCnt >= 2 && _sysExCnt <= 5)
            {
                int t = _sysExCnt - 2;
                _tone[t] = new byte[0x8b];
                for (int n = 0; n < 0x8b; n++)
                {
                    _tone[t][n] = msg[n + 0x0b];
                }
            }
            _sysExCnt++;
        }

        #region Patch Data
        private byte[] _common;
        private byte[] _tmt;
        private byte[][] _tone = { null, null, null, null };
        private int _sysExCnt = 0;
        private byte[][] _test = new byte[10][];
        #endregion

        #region Static data
        private static char[] blanks = { ' ' };

        private static string[] _categoryShort = {
            "---", "PNO", "EP", "KEY", "BEL", "MLT", "ORG", "ACD", 
            "HRM", "AGT", "EGT", "DGT", "BS", "SBS", "STR", "ORC", 
            "HIT", "WND", "FLT", "BRS", "SBR", "SAX", "HLD", "SLD", 
            "TEK", "PLS", "FX", "SYN", "BPD", "SPD", "VOX", "PLK", 
            "ETH", "FRT", "PRC", "SFX", "BTS", "DRM", "CMB"
        };

        private static string[] _categoryLong = {
            "No Assign", "AC.Piano", "EL.Piano", "Keyboards", "Bell", "Mallet", 
            "Organ", "Accordion", "Harmonica", "AC.Guitar", "EL.Guitar", "DIST.Guitar", 
            "Bass", "Synth Bass", "Strings", "Orchestra", "Hit&Stab", "Wind", "Flute", 
            "AC.Brass", "Synth Brass", "Sax", "Hard Lead", "Soft Lead", "Techno Synth", 
            "Pulsating", "Synth FX", "Other Synth", "Bright Pad", "Soft Pad", "Vox", 
            "Plucked", "Ethnic", "Fretted", "Percussion", "Sound FX", "Beat&Groove", 
            "Drums", "Combination"
        };

        #endregion
    }
}
