#region License
/* Copyright (c) 2007 Dirk Schnorpfeil
 * Midi Multimedia Libraries Copyright (c) 2005 Leslie Sanford
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to 
 * deal in the Software without restriction, including without limitation the 
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
 * sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
 * THE SOFTWARE.
 */
#endregion

#region Contact
/*
 * Dirk Schnorpfeil
 * Email: D.Schnorpfeil@web.de
 */
#endregion

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Diagnostics;

namespace MC808Librarian
{
    class JglImporter
    {
        static char[] separator = { '\t' };
        static char[] blanks = { ' ' };
        
        private byte[] _block = null;
        private int idx = 0;
        private List<Patch> _imported = new List<Patch>();
        private WaveConv _waves = new WaveConv();

        int GetInt4(int i)
        {
            int dw = 0;
            for (int n = 0; n < 4; n++)
            {
                dw = dw * 256;
                dw += _block[i + n];
            }
            return dw;
        }

        void WriteCard16(byte[,] _block, int t, int adr, int val)
        {
            byte d = (byte)(val & 0x0F);
            byte c = (byte)((val >> 4) & 0x0F);
            byte b = (byte)((val >> 8) & 0x0F);
            byte a = (byte)((val >> 12) & 0x0F);
            _block[t, adr] = a;
            _block[t, adr + 1] = b;
            _block[t, adr + 2] = c;
            _block[t, adr + 3] = d;
        }

        bool ImportBlock()
        {
            idx = 160;

            while (idx < _block.GetLength(0))
            {
                int ChunkSize = GetInt4(idx);
                if (ChunkSize != 0x440)
                {
                    idx += ChunkSize + 4;
                    continue;
                }

                idx = idx + 4;
                if (GetInt4(idx) != 0x4F)
                    return false;
                idx = idx + 4;
                byte[] common = new byte[0x51];
                for (int i = 0; i < 0x4F; i++)
                {
                    common[i] = _block[idx++];
                }
                common[0x4F] = 0;
                common[0x50] = 0;

                byte[] mfx = new byte[0x91];
                if (GetInt4(idx) != 0x91)
                    return false;
                idx = idx + 4;
                for (int i = 0; i < 0x91; i++)
                {
                    mfx[i] = _block[idx++];
                }

                byte[] chorus = new byte[0x54];
                if (GetInt4(idx) != 0x54)
                    return false;
                idx = idx + 4;
                for (int i = 0; i < 0x54; i++)
                {
                    chorus[i] = _block[idx++];
                }

                byte[] reverb = new byte[0x53];
                if (GetInt4(idx) != 0x53)
                    return false;
                idx = idx + 4;
                for (int i = 0; i < 0x53; i++)
                {
                    reverb[i] = _block[idx++];
                }

                byte[] tmt = new byte[0x29];
                if (GetInt4(idx) != 0x29)
                    return false;
                idx = idx + 4;
                for (int i = 0; i < 0x29; i++)
                {
                    tmt[i] = _block[idx++];
                }
                int[] tone_on = { 0, 0, 0, 0 };
                tone_on[0] = tmt[0x05];
                tone_on[1] = tmt[0x0E];
                tone_on[2] = tmt[0x17];
                tone_on[3] = tmt[0x20];

                byte[,] tone = new byte[4, 0x8B];
                for (int t = 0; t < 4; t++)
                {
                    if (GetInt4(idx) != 0x9a)
                        return false;
                    idx = idx + 4;
                    for (int i = 0; i < 0x88; i++)
                    {
                        tone[t, i] = _block[idx++];
                    }
                    tone[t, 0x89] = tone[t, 0x6D];
                    tone[t, 0x8a] = tone[t, 0x7B];
                    idx = idx + 18;
                    int wave_group = (tone[t, 0x28] << 12) | (tone[t, 0x29] << 8) | (tone[t, 0x2A] << 4) | (tone[t, 0x2B]);
                    int wave_left = (tone[t, 0x2C] << 12) | (tone[t, 0x2D] << 8) | (tone[t, 0x2E] << 4) | (tone[t, 0x2F]);
                    int wave_right = (tone[t, 0x30] << 12) | (tone[t, 0x31] << 8) | (tone[t, 0x32] << 4) | (tone[t, 0x33]);

                    if (wave_left > 0 && wave_left <= 1267)
                    {
                        int wavNo = _waves.WaveNo(wave_left);
                        WriteCard16(tone, t, 0x2C, wavNo);
                    }
                    if (wave_right > 0 && wave_right <= 1267)
                    {
                        int wavNo = _waves.WaveNo(wave_right);
                        WriteCard16(tone, t, 0x30, wavNo);
                    }
                }

                for (int t = 0; t < 4; t++)
                {
                    while (_block[idx] != 0)
                        idx++;
                    idx++;
                }

                Patch patch = new Patch(common, tmt, tone);
                _imported.Add(patch);
            }
            return true;
        }

        public List<Patch> Import(string path)
        {
            _imported.Clear();

            try
            {
                FileStream stream = new FileStream(path, FileMode.Open, FileAccess.Read);
                BinaryReader reader = new BinaryReader(stream);
                _block = reader.ReadBytes(10000000);
                reader.Close();
                stream.Close();
            }
            catch (Exception e)
            {
                MessageBox.Show("Error reading *.jgl file '" + path + "'. Reason: " + e.Message);
                return null;
            }

            if (ImportBlock())
                return _imported;
            else
                return null;
        }
    }
}
