#region License
/* Copyright (c) 2010 Dirk Schnorpfeil
 * Midi Multimedia Libraries Copyright (c) 2005 Leslie Sanford
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to 
 * deal in the Software without restriction, including without limitation the 
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
 * sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
 * THE SOFTWARE.
 */
#endregion

#region Contact
/*
 * Dirk Schnorpfeil
 * Email: D.Schnorpfeil@web.de
 */
#endregion

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.IO;
using System.Drawing;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using System.Diagnostics;
using Microsoft.Win32;

using Sanford.Multimedia.Midi;
using Sanford.Multimedia.Midi.UI;


namespace MC808Librarian
{
    public partial class MainForm : Form
    {
        private string _version = "1.4, September 2010";

        #region Member Vars
        private static List<Patch> _clipboard = new List<Patch>();
        private static int[] adr_map = new int[] { 0x1100, 0x1120, 0x1140, 0x1160, 0x1200, 0x1220, 0x1240, 0x1260, 0x1300, 0x1320, 0x1340, 0x1360, 0x1400, 0x1420, 0x1440, 0x1460 };

        private string _filename = "";
        private bool _changed = false;

        private OutputDevice _outDevice;
        private InputDevice _inDevice;
        private int _outDeviceID = -1;
        private int _inDeviceID = -1;
        private OutputDeviceDialog outDialog = new OutputDeviceDialog();
        private InputDeviceDialog inDialog = new InputDeviceDialog();
        private int _midiChannel = 0;
        private int _mc808DeviceID = 17;
        private Patch _rcvPatch = null;
        #endregion

        #region Form helpers
        public MainForm()
        {
            InitializeComponent();
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            bool leave = true;
            if (_changed)
            {
                if (MessageBox.Show("There are changed items. Really want to exit?", "Exit program", MessageBoxButtons.YesNo) != DialogResult.Yes)
                {
                    e.Cancel = true;
                    leave = false;
                }
            }
            if (leave)
            {
                if (_inDeviceID >= 0)
                    _inDevice.Close();
                if (_outDeviceID >= 0)
                    _outDevice.Close();
            }
        }

        private void MainForm_Load(object sender, EventArgs e)
        {
            object o = Registry.GetValue("HKEY_CURRENT_USER\\MC808Librarian", "InDeviceID", -1);
            if (o != null)
                _inDeviceID = (int)o;
            o = Registry.GetValue("HKEY_CURRENT_USER\\MC808Librarian", "OutDeviceID", -1);
            if (o != null)
                _outDeviceID = (int)o;
            o = Registry.GetValue("HKEY_CURRENT_USER\\MC808Librarian", "MidiChannel", 0);
            if (o != null)
                _midiChannel = (int)o;
            SetChannel(_midiChannel);
            o = Registry.GetValue("HKEY_CURRENT_USER\\MC808Librarian", "MC808DeviceID", 17);
            if (o != null)
                _mc808DeviceID = (int)o;

            toolStripStatusLabelChannel.Text = "Part = " + (_midiChannel + 1).ToString();
            toolStripStatusLabelDevice.Text = "Device ID = " + _mc808DeviceID.ToString();

            #region Midi Device Initialization
            if (_outDeviceID >= 0)
            {
                try
                {
                    MidiOutCaps caps = OutputDevice.GetDeviceCapabilities(_outDeviceID);
                    toolStripStatusLabelMidiOut.Text = "Midi Out = " + caps.name;
                    _outDevice = new OutputDevice(_outDeviceID);
                }
                catch (OutputDeviceException ex)
                {
                    MessageBox.Show("Could not open midi output device. Reason: " + ex.Message);
                    _outDeviceID = -1;
                    _outDevice = null;
                }
            }
            if (_inDeviceID >= 0)
            {
                try
                {
                    MidiInCaps caps = InputDevice.GetDeviceCapabilities(_inDeviceID);
                    toolStripStatusLabelMidiIn.Text = "Midi In = " + caps.name;
                    _inDevice = new InputDevice(_inDeviceID);
                    _inDevice.SysExMessageReceived += HandleSysExMessageReceived;
                }
                catch (InputDeviceException ex)
                {
                    MessageBox.Show("Could not open midi input device. Reason: " + ex.Message);
                    _inDeviceID = -1;
                    _inDevice = null;
                }
            }
            #endregion

            saveAsToolStripMenuItem.Enabled = false;
            saveToolStripMenuItem.Enabled = false;
            pasteToolStripMenuItem.Enabled = false;
            copyToolStripMenuItem.Enabled = false;
            deleteToolStripMenuItem.Enabled = false;
            sendPatchToolStripMenuItem.Enabled = false;
        }
        #endregion

        #region ListView helpers
        private void patches_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (patches.SelectedIndices.Count > 0)
            {
                copyToolStripMenuItem.Enabled = true;
                deleteToolStripMenuItem.Enabled = true;
                if (patches.SelectedIndices.Count == 1)
                    sendPatchToolStripMenuItem.Enabled = true;
                else
                    sendPatchToolStripMenuItem.Enabled = false;
            }
            else
            {
                copyToolStripMenuItem.Enabled = false;
                deleteToolStripMenuItem.Enabled = false;
                sendPatchToolStripMenuItem.Enabled = false;
            }
        }

        private void patches_ColumnClick(object sender, ColumnClickEventArgs e)
        {
            List<Patch> list = new List<Patch>();
            for (int n = 0; n < patches.Items.Count; n++)
            {
                Patch patch = patches.Items[n].Tag as Patch;
                list.Add(patch);
            }
            if (e.Column == 0)
                list.Sort(new Patch.ByName());
            else if (e.Column == 1)
                list.Sort(new Patch.ByCatShort());
            else if (e.Column == 2)
                list.Sort(new Patch.ByCatLong());

            patches.Items.Clear();
            for (int n = 0; n < list.Count; n++)
            {
                Patch patch = list[n];
                ListViewItem item = patches.Items.Add(patch.Name);
                item.SubItems.Add(patch.CategoryShort);
                item.SubItems.Add(patch.CategoryLong);
                item.Tag = patch;
            }
            _changed = true;
        }

        private void patches_DoubleClick(object sender, EventArgs e)
        {
            sendPatchToolStripMenuItem_Click(new object(), new EventArgs());
        }
        #endregion

        #region Library Menu
        private bool SaveFile(string path)
        {
            try
            {
                FileStream stream = new FileStream(path, FileMode.Create, FileAccess.Write);
                BinaryWriter writer = new BinaryWriter(stream);
                writer.Write((int)1);
                int count = patches.Items.Count;
                writer.Write(count);
                for (int n = 0; n < count; n++)
                {
                    ListViewItem item = patches.Items[n];
                    Patch patch = item.Tag as Patch;
                    writer.Write((int)0);    // type
                    patch.Write(writer);
                }
                writer.Close();
                stream.Close();
            }
            catch (Exception e)
            {
                MessageBox.Show("Cannot save to file '" + path + "'. Reason: " + e.Message);
                return false;
            }

            _changed = false;
            return true;
        }

        private bool LoadFile(string path)
        {
            patches.Items.Clear();
            _filename = "";

            try
            {
                FileStream stream = new FileStream(path, FileMode.Open, FileAccess.Read);
                BinaryReader reader = new BinaryReader(stream);
                int version = reader.ReadInt32();

                int count = reader.ReadInt32();
                for (int n = 0; n < count; n++)
                {
                    int type = reader.ReadInt32();
                    if (type == 0)
                    {
                        Patch patch = new Patch();
                        patch.Read(reader);
                        ListViewItem item = patches.Items.Add(patch.Name);
                        item.SubItems.Add(patch.CategoryShort);
                        item.SubItems.Add(patch.CategoryLong);
                        item.Tag = patch;
                    }
                }
                reader.Close();
                stream.Close();
            }
            catch (Exception e)
            {
                MessageBox.Show("Cannot load file '" + path + "'. Reason: " + e.Message);
                return false;
            }

            _filename = path;
            _changed = false;
            return true;
        }

        private void loadToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog dlg = new OpenFileDialog();

            dlg.Filter = "m8l files (*.m8l)|*.m8l|All files (*.*)|*.*";
            dlg.FilterIndex = 1;
            dlg.RestoreDirectory = true;

            if (dlg.ShowDialog() == DialogResult.OK)
            {
                if (LoadFile(dlg.FileName))
                {
                    saveAsToolStripMenuItem.Enabled = true;
                    saveToolStripMenuItem.Enabled = true;
                }
            }
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (_filename != "")
                SaveFile(_filename);
        }

        private void saveAsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SaveFileDialog dlg = new SaveFileDialog();
            dlg.Filter = "m8l files (*.m8l)|*.m8l|All files (*.*)|*.*";
            dlg.FilterIndex = 1;
            dlg.RestoreDirectory = true;

            if (dlg.ShowDialog() == DialogResult.OK)
            {
                SaveFile(dlg.FileName);
            }
        }

        private void importToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog dlg = new OpenFileDialog();

            dlg.Filter = "jgl files (*.jgl)|*.jgl|All files (*.*)|*.*";
            dlg.FilterIndex = 1;
            dlg.RestoreDirectory = true;

            if (dlg.ShowDialog() == DialogResult.OK)
            {
                JglImporter importer = new JglImporter();
                List<Patch> imported = importer.Import(dlg.FileName);
                if (imported == null)
                {
                    MessageBox.Show("Bad file format, no patches imported!");
                    return;
                }
                patches.Items.Clear();
                for (int n = 0; n < imported.Count; n++)
                {
                    Patch patch = imported[n];
                    ListViewItem item = patches.Items.Add(patch.Name);
                    item.SubItems.Add(patch.CategoryShort);
                    item.SubItems.Add(patch.CategoryLong);
                    item.Tag = patch;
                }
                saveAsToolStripMenuItem.Enabled = true;
                saveToolStripMenuItem.Enabled = true;
                _changed = true;
                _filename = "";
            }

        }

        private void newTtoolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (_changed)
            {
                if (MessageBox.Show("All changes will be lost. Really continue?", "New Library", MessageBoxButtons.YesNo) != DialogResult.Yes)
                    return;
            }

            patches.Items.Clear();

            _changed = false;
            _filename = "";
            saveAsToolStripMenuItem.Enabled = false;
            saveToolStripMenuItem.Enabled = false;
            copyToolStripMenuItem.Enabled = false;
            deleteToolStripMenuItem.Enabled = false;
            sendPatchToolStripMenuItem.Enabled = false;
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }
        #endregion

        #region Edit Menu
        private void copyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            _clipboard.Clear();
            for (int n = 0; n < patches.SelectedIndices.Count; n++)
            {
                Patch patch = patches.Items[patches.SelectedIndices[n]].Tag as Patch;
                _clipboard.Add(patch);
            }
            pasteToolStripMenuItem.Enabled = (_clipboard.Count > 0);
        }

        private void pasteToolStripMenuItem_Click(object sender, EventArgs e)
        {
            int idx = patches.SelectedIndices.Count == 0 ? 0 : patches.SelectedIndices[0];
            for (int n = 0; n < _clipboard.Count; n++)
            {
                Patch patch = _clipboard[n];
                ListViewItem item = patches.Items.Insert(idx + n, patch.Name);
                item.SubItems.Add(patch.CategoryShort);
                item.SubItems.Add(patch.CategoryLong);
                item.Tag = patch;
            }
            _changed = true;
        }

        private void deleteToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ListView.SelectedIndexCollection selected = patches.SelectedIndices;
            for (int n = selected.Count - 1; n >= 0; n--)
            {
                patches.Items.RemoveAt(selected[n]);
            }
            _changed = true;
        }
        #endregion

        #region Midi Menu
        private void HandleSysExMessageReceived(object sender, SysExMessageEventArgs e)
        {
            Debug.WriteLine(e.Message.Length.ToString());
            if (_rcvPatch != null)
            {
                // msg = new byte[e.Message.Length];
                byte[] msg = e.Message.GetBytes();
                _rcvPatch.AddSysex(e.Message.GetBytes());
            }
        }

        private void receivePatchToolStripMenuItem_Click(object sender, EventArgs e)
        {
            DoReceive();
        }

        private void buttonReceive_Click(object sender, EventArgs e)
        {
            DoReceive();
        }

        private void buttonSend_Click(object sender, EventArgs e)
        {
            DoSend();
        }
        
        private void sendPatchToolStripMenuItem_Click(object sender, EventArgs e)
        {
            DoSend();
        }
        
        private void DoReceive()
        {
            if (_rcvPatch != null )
                return;

            if (_inDevice == null)
            {
                MessageBox.Show("Please select a midi input device first!");
                return;
            }

            statusStrip.Text = "Receiving patch...";
            Application.DoEvents();

            try
            {
                _inDevice.Reset();
                _inDevice.AddSysExBuffer();
                _inDevice.AddSysExBuffer();
                _inDevice.AddSysExBuffer();
                _inDevice.AddSysExBuffer();
                _inDevice.AddSysExBuffer();
                _inDevice.AddSysExBuffer();
                _inDevice.AddSysExBuffer();
                _inDevice.StartRecording();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Stop);
                return;
            }

            _rcvPatch = new Patch();

            byte[] req = new byte[17];
            int adr = 0x1100 + _midiChannel * 0x20;
            req[0] = 0xF0; req[1] = 0x41; req[2] = (byte)(_mc808DeviceID - 1); req[3] = 0x00; req[4] = 0x00; req[5] = 0x14;
            req[6] = 0x11;  
            req[7] = (byte)(adr / 256); req[8] = (byte)(adr % 256); req[9] = 0x00; req[10] = 0x00;
            req[11] = 0x00; req[12] = 0x00; req[13] = 0x27; req[14] = 0x0B; req[16] = 0xF7;
            int sum = 0;
            for (int n = 7; n <= 14; n++)
                sum += req[n];
            sum = 128 - (sum % 128);
            req[15] = (byte)sum;

            SysExMessage msg = new SysExMessage(req);
            _outDevice.Send(msg);
            int iterations = 0;
            while (_rcvPatch.SysExCount < 6 && iterations < 99)
            {
                iterations++;
                Application.DoEvents();
                Thread.Sleep(50);
            }
            _inDevice.StopRecording();
            if (iterations >= 99)
            {
                MessageBox.Show("MC-808 did not answer. Abort.");
                _rcvPatch = null;
                return;
            }
            ListViewItem item = patches.Items.Insert(patches.Items.Count, _rcvPatch.Name);
            item.SubItems.Add(_rcvPatch.CategoryShort);
            item.SubItems.Add(_rcvPatch.CategoryLong);
            item.Tag = _rcvPatch;
            _rcvPatch = null;
            
            statusStrip.Text = "Ready.";
        }

        private void DoSend()
        {
            if (_outDevice == null)
            {
                MessageBox.Show("Please select a midi output device first!");
                return;
            }
            if (patches.SelectedItems.Count == 0)
                return;

            statusStrip.Text = "Sending patch...";
            this.Refresh();
            Patch patch = patches.SelectedItems[0].Tag as Patch;

            byte[] common = new byte[0x5E];
            int adr = 0x1100 + _midiChannel * 0x20;
            common[0] = 0xf0; common[1] = 0x41; common[2] = (byte)(_mc808DeviceID - 1); common[3] = 0x00;
            common[4] = 0x00; common[5] = 0x14; common[6] = 0x12; common[7] = (byte)(adr / 256);
            common[8] = (byte)(adr % 256); common[9] = 0x00; common[10] = 0x00;
            int idx = 11;
            int sum = common[7] + common[8];
            for (int i = 0; i < patch.Common.GetLength(0); i++)
            {
                common[idx++] = patch.Common[i];
                sum += patch.Common[i];
            }
            sum = sum % 128;
            sum = 128 - sum;
            common[idx++] = (byte)sum;
            common[idx] = 0xF7;

            SysExMessage msg = new SysExMessage(common);
            _outDevice.Send(msg);
            Thread.Sleep(100);

            byte[] tmt = new byte[0x37];
            tmt[0] = 0xf0; tmt[1] = 0x41; tmt[2] = (byte)(_mc808DeviceID - 1); tmt[3] = 0x00;
            tmt[4] = 0x00; tmt[5] = 0x14; tmt[6] = 0x12; tmt[7] = (byte)(adr / 256);
            tmt[8] = (byte)(adr % 256); tmt[9] = 0x10; tmt[10] = 0x00;
            idx = 11;
            sum = 0;
            for (int n = 7; n <= 10; n++)
                sum += tmt[n];;
            for (int i = 0; i < patch.Tmt.GetLength(0); i++)
            {
                tmt[idx++] = patch.Tmt[i];
                sum += patch.Tmt[i];
            }
            sum = sum % 128;
            sum = 128 - sum;
            tmt[idx++] = (byte)sum;
            tmt[idx] = 0xF7;
            msg = new SysExMessage(tmt);
            _outDevice.Send(msg);
            Thread.Sleep(100);

            for (int t = 0; t < 4; t++)
            {
                byte[] tone = new byte[0x99];
                tone[0] = 0xf0; tone[1] = 0x41; tone[2] = (byte)(_mc808DeviceID - 1); tone[3] = 0x00;
                tone[4] = 0x00; tone[5] = 0x14; tone[6] = 0x12; tone[7] = (byte)(adr / 256);
                tone[8] = (byte)(adr % 256); tone[9] = (byte)(0x20 + t * 2); tone[10] = 0x00;
                idx = 11;
                sum = 0;
                for (int n = 0; n <= 10; n++) sum += tone[n];
                for (int i = 0; i < patch.Tone(t).GetLength(0); i++)
                {
                    tone[idx++] = patch.Tone(t)[i];
                    sum += patch.Tone(t)[i];
                }
                sum = sum % 128;
                sum = 128 - sum;
                tone[idx++] = (byte)sum;
                tone[idx] = 0xF7;
                msg = new SysExMessage(tone);
                _outDevice.Send(msg);
                Thread.Sleep(100);
            }
            statusStrip.Text = "Ready.";
            this.Refresh();
        }
        #endregion

        #region Options menu
        #region Midi Channel

        private Button ButtonByNo(int channel)
        {
            switch (channel)
            {
                case 0: return button1;
                case 1: return button2;
                case 2: return button3;
                case 3: return button4;
                case 4: return button5;
                case 5: return button6;
                case 6: return button7;
                case 7: return button8;
                case 8: return button9;
                case 9: return button10;
                case 10: return button11;
                case 11: return button12;
                case 12: return button13;
                case 13: return button14;
                case 14: return button15;
                case 15: return button16;
                default: return null;
            }
        }

        private void SetChannel(int channel)
        {
            
            _midiChannel = channel;
            Registry.SetValue("HKEY_CURRENT_USER\\MC808Librarian", "MidiChannel", _midiChannel);
            toolStripStatusLabelChannel.Text = "Part = " + (_midiChannel + 1).ToString();
            for (int n = 0; n < 16; n++)
            {
                if (n != channel)
                    ButtonByNo(n).Enabled = true;
                else
                    ButtonByNo(n).Enabled = false;
            }
        }

        private void toolStripMenuItemCh1_Click(object sender, EventArgs e)
        {
            SetChannel(0);
        }

        private void toolStripMenuItemCh2_Click(object sender, EventArgs e)
        {
            SetChannel(1);
        }

        private void toolStripMenuItemCh3_Click(object sender, EventArgs e)
        {
            SetChannel(2);
        }

        private void toolStripMenuItemCh4_Click(object sender, EventArgs e)
        {
            SetChannel(3);
        }

        private void toolStripMenuItemCh5_Click(object sender, EventArgs e)
        {
            SetChannel(4);
        }

        private void toolStripMenuItem6_Click(object sender, EventArgs e)
        {
            SetChannel(5);
        }

        private void toolStripMenuItemCh7_Click(object sender, EventArgs e)
        {
            SetChannel(6);
        }

        private void toolStripMenuItemCh8_Click(object sender, EventArgs e)
        {
            SetChannel(7);
        }

        private void toolStripMenuItemCh9_Click(object sender, EventArgs e)
        {
            SetChannel(8);
        }

        private void toolStripMenuItemCh10_Click(object sender, EventArgs e)
        {
            SetChannel(9);
        }

        private void toolStripMenuItemCh11_Click(object sender, EventArgs e)
        {
            SetChannel(10);
        }

        private void toolStripMenuItemCh12_Click(object sender, EventArgs e)
        {
            SetChannel(11);
        }

        private void toolStripMenuItemCh13_Click(object sender, EventArgs e)
        {
            SetChannel(12);
        }

        private void toolStripMenuItemCh14_Click(object sender, EventArgs e)
        {
            SetChannel(13);
        }

        private void toolStripMenuItemCh15_Click(object sender, EventArgs e)
        {
            SetChannel(14);
        }

        private void toolStripMenuItemCh16_Click(object sender, EventArgs e)
        {
            SetChannel(15);
        }

        private void button1_Click(object sender, EventArgs e)
        {
            SetChannel(0);
        }

        private void button2_Click(object sender, EventArgs e)
        {
            SetChannel(1);
        }

        private void button3_Click(object sender, EventArgs e)
        {
            SetChannel(2);
        }

        private void button4_Click(object sender, EventArgs e)
        {
            SetChannel(3);
        }

        private void button5_Click(object sender, EventArgs e)
        {
            SetChannel(4);
        }

        private void button6_Click(object sender, EventArgs e)
        {
            SetChannel(5);
        }

        private void button7_Click(object sender, EventArgs e)
        {
            SetChannel(6);
        }

        private void button8_Click(object sender, EventArgs e)
        {
            SetChannel(7);
        }

        private void button9_Click(object sender, EventArgs e)
        {
            SetChannel(8);
        }

        private void button10_Click(object sender, EventArgs e)
        {
            SetChannel(9);
        }

        private void button11_Click(object sender, EventArgs e)
        {
            SetChannel(10);
        }

        private void button12_Click(object sender, EventArgs e)
        {
            SetChannel(11);
        }

        private void button13_Click(object sender, EventArgs e)
        {
            SetChannel(12);
        }

        private void button14_Click(object sender, EventArgs e)
        {
            SetChannel(13);
        }

        private void button15_Click(object sender, EventArgs e)
        {
            SetChannel(14);
        }

        private void button16_Click(object sender, EventArgs e)
        {
            SetChannel(15);
        }
        
        #endregion

        #region MC808 Device ID
        private void Set808DeviceID(int deviceID)
        {
            _mc808DeviceID = deviceID;
            Registry.SetValue("HKEY_CURRENT_USER\\MC808Librarian", "MC808DeviceID", _mc808DeviceID);
            toolStripStatusLabelDevice.Text = "Device ID = " + _mc808DeviceID.ToString();
        }

        private void toolStripMenuItemDev17_Click(object sender, EventArgs e)
        {
            Set808DeviceID(17);
        }

        private void toolStripMenuItemDev18_Click(object sender, EventArgs e)
        {
            Set808DeviceID(18);
        }

        private void toolStripMenuItemDev19_Click(object sender, EventArgs e)
        {
            Set808DeviceID(19);
        }

        private void toolStripMenuItemDev20_Click(object sender, EventArgs e)
        {
            Set808DeviceID(20);
        }

        private void toolStripMenuItemDev21_Click(object sender, EventArgs e)
        {
            Set808DeviceID(21);
        }

        private void toolStripMenuItemDev22_Click(object sender, EventArgs e)
        {
            Set808DeviceID(22);
        }

        private void toolStripMenuItemDev23_Click(object sender, EventArgs e)
        {
            Set808DeviceID(23);
        }

        private void toolStripMenuItemDev24_Click(object sender, EventArgs e)
        {
            Set808DeviceID(24);
        }

        private void toolStripMenuItemDev25_Click(object sender, EventArgs e)
        {
            Set808DeviceID(25);
        }

        private void toolStripMenuItemDev26_Click(object sender, EventArgs e)
        {
            Set808DeviceID(26);
        }

        private void toolStripMenuItemDev27_Click(object sender, EventArgs e)
        {
            Set808DeviceID(27);
        }

        private void toolStripMenuItemDev28_Click(object sender, EventArgs e)
        {
            Set808DeviceID(28);
        }

        private void toolStripMenuItemDev29_Click(object sender, EventArgs e)
        {
            Set808DeviceID(29);
        }

        private void toolStripMenuItemDev30_Click(object sender, EventArgs e)
        {
            Set808DeviceID(30);
        }

        private void toolStripMenuItemDev31_Click(object sender, EventArgs e)
        {
            Set808DeviceID(31);
        }

        private void toolStripMenuItemDev32_Click(object sender, EventArgs e)
        {
            Set808DeviceID(32);
        }
        #endregion

        private void midiInToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (inDialog.ShowDialog() == DialogResult.OK)
            {
                if (_inDeviceID >= 0)
                    _inDevice.Close();
                try
                {
                    if (inDialog.InputDeviceID >= 0)
                    {
                        _inDeviceID = inDialog.InputDeviceID;
                        _inDevice = new InputDevice(_inDeviceID);
                        _inDevice.SysExMessageReceived += HandleSysExMessageReceived;
                    }
                }
                catch (Exception)
                {
                    _inDeviceID = -1;
                }
                Registry.SetValue("HKEY_CURRENT_USER\\MC808Librarian", "InDeviceID", _inDeviceID);
                if (_inDeviceID >= 0)
                {
                    MidiInCaps caps = InputDevice.GetDeviceCapabilities(_inDeviceID);
                    toolStripStatusLabelMidiIn.Text = "Midi In = " + caps.name;
                }
                else
                    toolStripStatusLabelMidiIn.Text = "Midi In = ???";
            }
        }

        private void midiOutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (outDialog.ShowDialog() == DialogResult.OK)
            {
                if (_outDeviceID >= 0)
                    _outDevice.Close();
                try
                {
                    if (outDialog.OutputDeviceID >= 0)
                    {
                        _outDeviceID = outDialog.OutputDeviceID;
                        _outDevice = new OutputDevice(_outDeviceID);
                    }
                }
                catch (Exception)
                {
                    _outDeviceID = -1;
                }
                Registry.SetValue("HKEY_CURRENT_USER\\MC808Librarian", "OutDeviceID", _outDeviceID);
                if (_outDeviceID >= 0)
                {
                    MidiOutCaps caps = OutputDevice.GetDeviceCapabilities(_outDeviceID);
                    toolStripStatusLabelMidiOut.Text = "Midi Out = " + caps.name;
                }
                else
                    toolStripStatusLabelMidiOut.Text = "Midi Out = ???";
            }
        }
        #endregion

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            string txt =
                 "Version: " + _version + "\n\n" +
                 "Copyright (c) 2010 Dirk Schnorpfeil\n\nMidi Multimedia Libraries Copyright (c) 2005 Leslie Sanford (http://www.codeproject.com/cs/media/MIDIToolkit.asp)\n" +
                 "\n" +
                 "Permission is hereby granted, free of charge, to any person obtaining a copy " +
                 "of this software and associated documentation files (the 'Software'), to " +
                 "deal in the Software without restriction, including without limitation the " +
                 "rights to use, copy, modify, merge, publish, distribute, sublicense, and/or " +
                 "sell copies of the Software, and to permit persons to whom the Software is " +
                 "furnished to do so, subject to the following conditions:\n" +
                 "\n" +
                 "The above copyright notice and this permission notice shall be included in " +
                 "all copies or substantial portions of the Software.\n" +
                 "\n" +
                 "THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR " +
                 "IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, " +
                 "FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE " +
                 "AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER " +
                 "LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, " +
                 "OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN " +
                 "THE SOFTWARE.\n\nContact: 'D.Schnorpfeil@web.de'";

            MessageBox.Show(txt, "About MC-808 Librarian");
        }

        #region Midi File Export
        private void exportPatchesAsMidiFilesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            DoWriteMidiFile();
        }


        private void WriteVariableLengthValue(Stream stream, int value)
        {
            int num = value;
            byte[] buffer = new byte[0x4];
            int index = 0x0;
            buffer[0x0] = (byte)(num & 0x7f);
            for (num = num >> 0x7; num > 0x0; num = num >> 0x7)
            {
                index++;
                buffer[index] = (byte)((num & 0x7f) | 0x80);
            }
            while (index >= 0x0)
            {
                stream.WriteByte(buffer[index]);
                index--;
            }
        }

        private void DoWriteMidiFile()
        {
            if (this.patches.SelectedItems.Count != 0x0)
            {
                FolderBrowserDialog dialog = new FolderBrowserDialog();
                dialog.Description = "Select a directory for the midi file(s)";
                if (dialog.ShowDialog() == DialogResult.OK)
                {
                    string selectedPath = dialog.SelectedPath;
                    for (int i = 0x0; i < this.patches.SelectedIndices.Count; i++)
                    {
                        Patch tag = this.patches.Items[this.patches.SelectedIndices[i]].Tag as Patch;
                        string str2 = tag.Name.Replace('>', '_').Replace('<', '_').Replace('|', '_').Replace('+', '_').Replace('\\', '_').Replace('/', '_').Replace(':', '_').Replace('.', '_').Replace('?', '_').Replace('*', '_') + ".mid";
                        FileStream stream = null;
                        try
                        {
                            stream = new FileStream(selectedPath + @"\" + str2, FileMode.Create);
                        }
                        catch (Exception exception)
                        {
                            MessageBox.Show("Cannot write midi file '" + selectedPath + @"\" + str2 + "' for reason: " + exception.Message + ". Abort.");
                            break;
                        }
                        stream.WriteByte(0x4d);
                        stream.WriteByte(0x54);
                        stream.WriteByte(0x68);
                        stream.WriteByte(0x64);
                        stream.WriteByte(0x0);
                        stream.WriteByte(0x0);
                        stream.WriteByte(0x0);
                        stream.WriteByte(0x6);
                        stream.WriteByte(0x0);
                        stream.WriteByte(0x0);
                        stream.WriteByte(0x0);
                        stream.WriteByte(0x1);
                        stream.WriteByte(0x0);
                        stream.WriteByte(0x60);
                        stream.WriteByte(0x4d);
                        stream.WriteByte(0x54);
                        stream.WriteByte(0x72);
                        stream.WriteByte(0x6b);
                        byte[] bytes = BitConverter.GetBytes(0x308);
                        if (BitConverter.IsLittleEndian)
                        {
                            Array.Reverse(bytes);
                        }
                        stream.Write(bytes, 0x0, bytes.Length);
                        byte[] buffer2 = new byte[0x5e];
                        int num2 = adr_map[this._midiChannel];
                        buffer2[0x0] = 0xf0;
                        buffer2[0x1] = 0x41;
                        buffer2[0x2] = (byte)(this._mc808DeviceID - 0x1);
                        buffer2[0x3] = 0x0;
                        buffer2[0x4] = 0x0;
                        buffer2[0x5] = 0x14;
                        buffer2[0x6] = 0x12;
                        buffer2[0x7] = (byte)(num2 / 0x100);
                        buffer2[0x8] = (byte)(num2 % 0x100);
                        buffer2[0x9] = 0x0;
                        buffer2[0xa] = 0x0;
                        int index = 0xb;
                        int num4 = buffer2[0x7] + buffer2[0x8];
                        for (int j = 0x0; j < tag.Common.GetLength(0x0); j++)
                        {
                            buffer2[index++] = tag.Common[j];
                            num4 += tag.Common[j];
                        }
                        num4 = num4 % 0x80;
                        num4 = 0x80 - num4;
                        buffer2[index++] = (byte)num4;
                        buffer2[index] = 0xf7;
                        stream.WriteByte(0xc);
                        stream.WriteByte(0xf0);
                        this.WriteVariableLengthValue(stream, index);
                        for (int k = 0x1; k <= index; k++)
                        {
                            stream.WriteByte(buffer2[k]);
                        }
                        byte[] buffer3 = new byte[0x37];
                        buffer3[0x0] = 0xf0;
                        buffer3[0x1] = 0x41;
                        buffer3[0x2] = (byte)(this._mc808DeviceID - 0x1);
                        buffer3[0x3] = 0x0;
                        buffer3[0x4] = 0x0;
                        buffer3[0x5] = 0x14;
                        buffer3[0x6] = 0x12;
                        buffer3[0x7] = (byte)(num2 / 0x100);
                        buffer3[0x8] = (byte)(num2 % 0x100);
                        buffer3[0x9] = 0x10;
                        buffer3[0xa] = 0x0;
                        index = 0xb;
                        num4 = 0x0;
                        for (int m = 0x7; m <= 0xa; m++)
                        {
                            num4 += buffer3[m];
                        }
                        for (int n = 0x0; n < tag.Tmt.GetLength(0x0); n++)
                        {
                            buffer3[index++] = tag.Tmt[n];
                            num4 += tag.Tmt[n];
                        }
                        num4 = num4 % 0x80;
                        num4 = 0x80 - num4;
                        buffer3[index++] = (byte)num4;
                        buffer3[index] = 0xf7;
                        stream.WriteByte(0xc);
                        stream.WriteByte(0xf0);
                        this.WriteVariableLengthValue(stream, index);
                        for (int num9 = 0x1; num9 <= index; num9++)
                        {
                            stream.WriteByte(buffer3[num9]);
                        }
                        for (int num10 = 0x0; num10 < 0x4; num10++)
                        {
                            byte[] buffer4 = new byte[0x99];
                            buffer4[0x0] = 0xf0;
                            buffer4[0x1] = 0x41;
                            buffer4[0x2] = (byte)(this._mc808DeviceID - 0x1);
                            buffer4[0x3] = 0x0;
                            buffer4[0x4] = 0x0;
                            buffer4[0x5] = 0x14;
                            buffer4[0x6] = 0x12;
                            buffer4[0x7] = (byte)(num2 / 0x100);
                            buffer4[0x8] = (byte)(num2 % 0x100);
                            buffer4[0x9] = (byte)(0x20 + (num10 * 0x2));
                            buffer4[0xa] = 0x0;
                            index = 0xb;
                            num4 = 0x0;
                            for (int num11 = 0x0; num11 <= 0xa; num11++)
                            {
                                num4 += buffer4[num11];
                            }
                            for (int num12 = 0x0; num12 < tag.Tone(num10).GetLength(0x0); num12++)
                            {
                                buffer4[index++] = tag.Tone(num10)[num12];
                                num4 += tag.Tone(num10)[num12];
                            }
                            num4 = num4 % 0x80;
                            num4 = 0x80 - num4;
                            buffer4[index++] = (byte)num4;
                            buffer4[index] = 0xf7;
                            stream.WriteByte(0xc);
                            stream.WriteByte(0xf0);
                            this.WriteVariableLengthValue(stream, index);
                            for (int num13 = 0x1; num13 <= index; num13++)
                            {
                                stream.WriteByte(buffer4[num13]);
                            }
                        }
                        stream.WriteByte(0x0);
                        stream.WriteByte(0xff);
                        stream.WriteByte(0x2f);
                        stream.WriteByte(0x0);
                        stream.Close();
                    }
                }
            }
        }
        #endregion
    }
}